@@ -1,3 +1,4 @@
+pub mod gen_stack;
 mod internal_states;
 mod snippets;
 
@@ -5,23 +6,26 @@ use std::cell::RefCell;
 use std::collections::{BTreeMap, HashMap, HashSet};
 use std::fs;
 use std::path::Path;
+use std::sync::Arc;
+
+use cirru_edn::EdnKwd;
+use im_ternary_tree::TernaryTreeList;
 
 use crate::builtins::meta::{js_gensym, reset_js_gensym_index};
-use crate::builtins::{is_js_syntax_procs, is_proc_name, is_syntax_name};
-use crate::call_stack;
+use crate::builtins::{is_js_syntax_procs, is_proc_name};
 use crate::call_stack::StackKind;
 use crate::primes;
-use crate::primes::{Calcit, CalcitItems, ImportRule, SymbolResolved::*};
+use crate::primes::{Calcit, CalcitItems, CalcitSyntax, ImportRule, SymbolResolved::*};
 use crate::program;
 use crate::util::string::{has_ns_part, matches_digits, matches_js_var, wrap_js_str};
 
-type ImportsDict = BTreeMap<String, ImportedTarget>;
+type ImportsDict = BTreeMap<Arc<str>, ImportedTarget>;
 
 #[derive(Debug, PartialEq, Clone)]
 pub enum ImportedTarget {
-  AsNs(String),
-  DefaultNs(String),
-  ReferNs(String),
+  AsNs(Arc<str>),
+  DefaultNs(Arc<str>),
+  ReferNs(Arc<str>),
 }
 
 fn to_js_import_name(ns: &str, mjs_mode: bool) -> String {
@@ -153,9 +157,9 @@ fn escape_cirru_str(s: &str) -> String {
   result
 }
 
-fn quote_to_js(xs: &Calcit, var_prefix: &str, keywords: &RefCell<Vec<String>>) -> Result<String, String> {
+fn quote_to_js(xs: &Calcit, var_prefix: &str, keywords: &RefCell<Vec<EdnKwd>>) -> Result<String, String> {
   match xs {
-    Calcit::Symbol(s, ..) => Ok(format!("new {}CalcitSymbol({})", var_prefix, escape_cirru_str(s))),
+    Calcit::Symbol { sym, .. } => Ok(format!("new {}CalcitSymbol({})", var_prefix, escape_cirru_str(sym))),
     Calcit::Str(s) => Ok(escape_cirru_str(s)),
     Calcit::Bool(b) => Ok(b.to_string()),
     Calcit::Number(n) => Ok(n.to_string()),
@@ -170,12 +174,12 @@ fn quote_to_js(xs: &Calcit, var_prefix: &str, keywords: &RefCell<Vec<String>>) -
         }
         chunk.push_str(&quote_to_js(y, var_prefix, keywords)?);
       }
-      Ok(format!("new {}CalcitList([{}])", var_prefix, chunk))
+      Ok(format!("new {}CalcitSliceList([{}])", var_prefix, chunk))
     }
     Calcit::Keyword(s) => {
       let mut kwds = keywords.borrow_mut();
       kwds.push(s.to_owned());
-      Ok(format!("_kwd[{}]", escape_cirru_str(s)))
+      Ok(format!("_kwd[{}]", escape_cirru_str(&s.to_string())))
     }
     _ => unreachable!(format!("Unexpected data in quote for js: {}", xs)),
   }
@@ -196,24 +200,26 @@ fn make_fn_wrapper(body: &str) -> String {
 fn to_js_code(
   xs: &Calcit,
   ns: &str,
-  local_defs: &HashSet<String>,
+  local_defs: &HashSet<Arc<str>>,
   file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<String>>,
+  keywords: &RefCell<Vec<EdnKwd>>,
   return_label: &Option<String>,
 ) -> Result<String, String> {
   if let Calcit::List(ys) = xs {
     gen_call_code(ys, ns, local_defs, xs, file_imports, keywords, return_label)
   } else {
     let ret = match xs {
-      Calcit::Symbol(s, def_ns, at_def, resolved) => {
-        gen_symbol_code(s, def_ns, at_def, resolved, ns, xs, local_defs, file_imports)
+      Calcit::Symbol {
+        sym,
+        ns: def_ns,
+        at_def,
+        resolved,
+      } => {
+        let resolved_info = resolved.to_owned().map(|v| (*v).to_owned());
+        gen_symbol_code(sym, &**def_ns, &**at_def, &resolved_info, ns, xs, local_defs, file_imports)
       }
       Calcit::Proc(s) => {
-        let proc_prefix = if ns == primes::CORE_NS {
-          "$calcit_procs."
-        } else {
-          "$calcit."
-        };
+        let proc_prefix = get_proc_prefix(ns);
         // println!("gen proc {} under {}", s, ns,);
         // let resolved = Some(ResolvedDef(String::from(primes::CORE_NS), s.to_owned()));
         // gen_symbol_code(s, primes::CORE_NS, &resolved, ns, xs, local_defs)
@@ -234,17 +240,8 @@ fn to_js_code(
         }
       }
       Calcit::Syntax(s, ..) => {
-        let resolved = Some(ResolvedDef(String::from(primes::CORE_NS), s.to_owned(), None));
-        gen_symbol_code(
-          s,
-          primes::CORE_NS,
-          primes::GENERATED_DEF,
-          &resolved,
-          ns,
-          xs,
-          local_defs,
-          file_imports,
-        )
+        let proc_prefix = get_proc_prefix(ns);
+        Ok(format!("{}{}", proc_prefix, escape_var(&s.to_string())))
       }
       Calcit::Str(s) => Ok(escape_cirru_str(s)),
       Calcit::Bool(b) => Ok(b.to_string()),
@@ -253,7 +250,7 @@ fn to_js_code(
       Calcit::Keyword(s) => {
         let mut kwds = keywords.borrow_mut();
         kwds.push(s.to_owned());
-        Ok(format!("_kwd[{}]", wrap_js_str(s)))
+        Ok(format!("_kwd[{}]", wrap_js_str(&s.to_string())))
       }
       Calcit::List(_) => unreachable!("[Error] list handled in another branch"),
       a => unreachable!(format!("[Error] unknown kind to gen js code: {}", a)),
@@ -269,10 +266,10 @@ fn to_js_code(
 fn gen_call_code(
   ys: &CalcitItems,
   ns: &str,
-  local_defs: &HashSet<String>,
+  local_defs: &HashSet<Arc<str>>,
   xs: &Calcit,
   file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<String>>,
+  keywords: &RefCell<Vec<EdnKwd>>,
   return_label: &Option<String>,
 ) -> Result<String, String> {
   let return_code = match &return_label {
@@ -280,25 +277,21 @@ fn gen_call_code(
     None => String::from(""),
   };
   let var_prefix = if ns == primes::CORE_NS { "" } else { "$calcit." };
-  let proc_prefix = if ns == primes::CORE_NS {
-    "$calcit_procs."
-  } else {
-    "$calcit."
-  };
+  let proc_prefix = get_proc_prefix(ns);
   if ys.is_empty() {
     println!("[Warn] Unexpected empty list inside {}", xs);
     return Ok(String::from("()"));
   }
 
   let head = ys[0].to_owned();
-  let body = ys.skip(1);
+  let body = ys.skip(1)?;
   match &head {
-    Calcit::Symbol(s, ..) | Calcit::Proc(s) | Calcit::Syntax(s, ..) => {
-      match s.as_str() {
-        "if" => {
+    Calcit::Syntax(s, ..) => {
+      match s {
+        CalcitSyntax::If => {
           if let Some(Calcit::List(ys)) = body.get(2) {
             if let Some(Calcit::Syntax(syn, ..)) = ys.get(0) {
-              if syn == "if" {
+              if syn == &CalcitSyntax::If {
                 return gen_if_code(&body, local_defs, xs, ns, file_imports, keywords, return_label);
               }
             }
@@ -308,38 +301,34 @@ fn gen_call_code(
           }
           return match (body.get(0), body.get(1)) {
             (Some(condition), Some(true_branch)) => {
-              call_stack::push_call_stack(ns, "if", StackKind::Codegen, xs.to_owned(), &im::vector![]);
+              gen_stack::push_call_stack(ns, "if", StackKind::Codegen, xs.to_owned(), &TernaryTreeList::Empty);
               let false_code = match body.get(2) {
                 Some(fal) => to_js_code(fal, ns, local_defs, file_imports, keywords, &None)?,
                 None => String::from("null"),
               };
               let cond_code = to_js_code(condition, ns, local_defs, file_imports, keywords, &None)?;
               let true_code = to_js_code(true_branch, ns, local_defs, file_imports, keywords, &None)?;
-              call_stack::pop_call_stack();
-              Ok(format!(
-                "{}( {} ? {} : {} )",
-                return_code, cond_code, true_code, false_code
-              ))
+              gen_stack::pop_call_stack();
+              Ok(format!("{}( {} ? {} : {} )", return_code, cond_code, true_code, false_code))
             }
             (_, _) => Err(format!("if expected 2~3 nodes, got: {:?}", body)),
           };
         }
-        "&let" => gen_let_code(&body, local_defs, xs, ns, file_imports, keywords, return_label),
-        ";" => Ok(format!("(/* {} */ null)", Calcit::List(body))),
+        CalcitSyntax::CoreLet => gen_let_code(&body, local_defs, xs, ns, file_imports, keywords, return_label),
 
-        "quote" => match body.get(0) {
+        CalcitSyntax::Quote => match body.get(0) {
           Some(item) => quote_to_js(item, var_prefix, keywords),
           None => Err(format!("quote expected a node, got nothing from {:?}", body)),
         },
-        "defatom" => {
+        CalcitSyntax::Defatom => {
           match (body.get(0), body.get(1)) {
             _ if body.len() > 2 => Err(format!("defatom expected name and value, got too many: {:?}", body)),
-            (Some(Calcit::Symbol(sym, ..)), Some(v)) => {
+            (Some(Calcit::Symbol { sym, .. }), Some(v)) => {
               // let _name = escape_var(sym); // TODO
               let ref_path = wrap_js_str(&format!("{}/{}", ns, sym.to_owned()));
-              call_stack::push_call_stack(ns, sym, StackKind::Codegen, xs.to_owned(), &im::vector![]);
+              gen_stack::push_call_stack(ns, sym, StackKind::Codegen, xs.to_owned(), &TernaryTreeList::Empty);
               let value_code = &to_js_code(v, ns, local_defs, file_imports, keywords, &None)?;
-              call_stack::pop_call_stack();
+              gen_stack::pop_call_stack();
               Ok(format!(
                 "\n({}peekDefatom({}) ?? {}defatom({}, {}))\n",
                 &var_prefix, &ref_path, &var_prefix, &ref_path, value_code
@@ -349,12 +338,12 @@ fn gen_call_code(
           }
         }
 
-        "defn" => match (body.get(0), body.get(1)) {
-          (Some(Calcit::Symbol(sym, ..)), Some(Calcit::List(ys))) => {
-            let func_body = body.skip(2);
-            call_stack::push_call_stack(ns, sym, StackKind::Codegen, xs.to_owned(), &im::vector![]);
+        CalcitSyntax::Defn => match (body.get(0), body.get(1)) {
+          (Some(Calcit::Symbol { sym, .. }), Some(Calcit::List(ys))) => {
+            let func_body = body.skip(2)?;
+            gen_stack::push_call_stack(ns, sym, StackKind::Codegen, xs.to_owned(), &TernaryTreeList::Empty);
             let ret = gen_js_func(sym, ys, &func_body, ns, false, local_defs, file_imports, keywords);
-            call_stack::pop_call_stack();
+            gen_stack::pop_call_stack();
             match ret {
               Ok(code) => Ok(format!("{}{}", return_code, code)),
               _ => ret,
@@ -363,8 +352,42 @@ fn gen_call_code(
           (_, _) => Err(format!("defn expected name arguments, got: {:?}", body)),
         },
 
-        "defmacro" => Ok(format!("/* Unexpected macro {} */", xs)),
-        "quasiquote" => Ok(format!("(/* Unexpected quasiquote {} */ null)", xs.lisp_str())),
+        CalcitSyntax::Defmacro => Ok(format!("/* Unexpected macro {} */", xs)),
+        CalcitSyntax::Quasiquote => Ok(format!("(/* Unexpected quasiquote {} */ null)", xs.lisp_str())),
+        CalcitSyntax::Try => match (body.get(0), body.get(1)) {
+          (Some(expr), Some(handler)) => {
+            gen_stack::push_call_stack(ns, "try", StackKind::Codegen, xs.to_owned(), &TernaryTreeList::Empty);
+            let next_return_label = match return_label {
+              Some(x) => Some(x.to_owned()),
+              None => Some(String::from("return ")),
+            };
+            let try_code = to_js_code(expr, ns, local_defs, file_imports, keywords, &next_return_label)?;
+            let err_var = js_gensym("errMsg");
+            let handler = to_js_code(handler, ns, local_defs, file_imports, keywords, &None)?;
+
+            gen_stack::pop_call_stack();
+            let code = snippets::tmpl_try(err_var, try_code, handler, &next_return_label.unwrap());
+            match return_label {
+              Some(_) => Ok(code),
+              None => Ok(snippets::tmpl_fn_wrapper(code)),
+            }
+          }
+          (_, _) => Err(format!("try expected 2 nodes, got {:?}", body)),
+        },
+        _ => {
+          let args_code = gen_args_code(&body, ns, local_defs, file_imports, keywords)?;
+          Ok(format!(
+            "{}{}({})",
+            return_code,
+            to_js_code(&head, ns, local_defs, file_imports, keywords, &None)?,
+            args_code
+          ))
+        }
+      }
+    }
+    Calcit::Symbol { sym: s, .. } | Calcit::Proc(s) => {
+      match &**s {
+        ";" => Ok(format!("(/* {} */ null)", Calcit::List(body))),
 
         "raise" => {
           // not core syntax, but treat as macro for better debugging experience
@@ -389,36 +412,17 @@ fn gen_call_code(
             None => Err(format!("raise expected 1~2 arguments, got {:?}", body)),
           }
         }
-        "try" => match (body.get(0), body.get(1)) {
-          (Some(expr), Some(handler)) => {
-            call_stack::push_call_stack(ns, "try", StackKind::Codegen, xs.to_owned(), &im::vector![]);
-            let next_return_label = match return_label {
-              Some(x) => Some(x.to_owned()),
-              None => Some(String::from("return ")),
-            };
-            let try_code = to_js_code(expr, ns, local_defs, file_imports, keywords, &next_return_label)?;
-            let err_var = js_gensym("errMsg");
-            let handler = to_js_code(handler, ns, local_defs, file_imports, keywords, &None)?;
 
-            call_stack::pop_call_stack();
-            let code = snippets::tmpl_try(err_var, try_code, handler, &next_return_label.unwrap());
-            match return_label {
-              Some(_) => Ok(code),
-              None => Ok(snippets::tmpl_fn_wrapper(code)),
-            }
-          }
-          (_, _) => Err(format!("try expected 2 nodes, got {:?}", body)),
-        },
         "echo" | "println" => {
           // not core syntax, but treat as macro for better debugging experience
-          let args = ys.skip(1);
+          let args = ys.skip(1)?;
           let args_code = gen_args_code(&args, ns, local_defs, file_imports, keywords)?;
           Ok(format!("console.log({}printable({}))", proc_prefix, args_code))
         }
         "exists?" => {
           // not core syntax, but treat as macro for availability
           match body.get(0) {
-            Some(Calcit::Symbol(_sym, ..)) => {
+            Some(Calcit::Symbol { .. }) => {
               let target = to_js_code(&body[0], ns, local_defs, file_imports, keywords, &None)?; // TODO could be simpler
               return Ok(format!("{}(typeof {} !== 'undefined')", return_code, target));
             }
@@ -428,7 +432,7 @@ fn gen_call_code(
         }
         "new" => match body.get(0) {
           Some(ctor) => {
-            let args = body.skip(1);
+            let args = body.skip(1)?;
             let args_code = gen_args_code(&args, ns, local_defs, file_imports, keywords)?;
             Ok(format!(
               "{}new {}({})",
@@ -506,7 +510,7 @@ fn gen_call_code(
           if matches_js_var(name) {
             match body.get(0) {
               Some(obj) => {
-                let args = body.skip(1);
+                let args = body.skip(1)?;
                 let args_code = gen_args_code(&args, ns, local_defs, file_imports, keywords)?;
                 Ok(format!(
                   "{}{}.{}({})",
@@ -526,7 +530,7 @@ fn gen_call_code(
           let name = s.strip_prefix('.').unwrap();
           match body.get(0) {
             Some(obj) => {
-              let args = body.skip(1);
+              let args = body.skip(1)?;
               let args_code = gen_args_code(&args, ns, local_defs, file_imports, keywords)?;
               Ok(format!(
                 "{}{}invoke_method({})({},{})",
@@ -571,25 +575,29 @@ fn gen_symbol_code(
   resolved: &Option<primes::SymbolResolved>,
   ns: &str,
   xs: &Calcit,
-  local_defs: &HashSet<String>,
+  local_defs: &HashSet<Arc<str>>,
   file_imports: &RefCell<ImportsDict>,
 ) -> Result<String, String> {
   // println!("gen symbol: {} {} {} {:?}", s, def_ns, ns, resolved);
-  let var_prefix = if ns == "calcit.core" { "" } else { "$calcit." };
+  let var_prefix = if ns == primes::CORE_NS { "" } else { "$calcit." };
   if has_ns_part(s) {
-    let ns_part = s.split('/').collect::<Vec<&str>>()[0]; // TODO
-    if ns_part == "js" {
+    if s.starts_with("js/") {
       Ok(escape_ns_var(s, "js"))
     } else {
-      // TODO ditry code
+      let ns_part = s.split('/').collect::<Vec<&str>>()[0];
+      // TODO dirty code
       // TODO namespace part supposed be parsed during preprocessing, this mimics old behaviors
       match resolved {
-        Some(ResolvedDef(r_ns, _r_def, _import_rule /* None */)) => {
+        Some(ResolvedDef {
+          ns: r_ns,
+          def: _r_def,
+          rule: _import_rule, /* None */
+        }) => {
           if is_cirru_string(r_ns) {
-            track_ns_import(ns_part.to_owned(), ImportedTarget::AsNs(r_ns.to_owned()), file_imports)?;
+            track_ns_import(ns_part, ImportedTarget::AsNs(r_ns.to_owned()), file_imports)?;
             Ok(escape_ns_var(s, ns_part))
           } else {
-            track_ns_import(r_ns.to_owned(), ImportedTarget::AsNs(r_ns.to_owned()), file_imports)?;
+            track_ns_import(r_ns, ImportedTarget::AsNs(r_ns.to_owned()), file_imports)?;
             Ok(escape_ns_var(s, r_ns))
           }
         }
@@ -598,26 +606,27 @@ fn gen_symbol_code(
         None => Err(format!("expected symbol with ns being resolved: {:?}", xs)),
       }
     }
-  } else if is_js_syntax_procs(s) || is_proc_name(s) || is_syntax_name(s) {
+  } else if is_js_syntax_procs(s) || is_proc_name(s) || CalcitSyntax::is_core_syntax(s) {
     // return Ok(format!("{}{}", var_prefix, escape_var(s)));
-    let proc_prefix = if ns == primes::CORE_NS {
-      "$calcit_procs."
-    } else {
-      "$calcit."
-    };
+    let proc_prefix = get_proc_prefix(ns);
     return Ok(format!("{}{}", proc_prefix, escape_var(s)));
   } else if matches!(resolved, Some(ResolvedLocal)) || local_defs.contains(s) {
     Ok(escape_var(s))
-  } else if let Some(ResolvedDef(r_ns, _r_def, import_rule)) = resolved.to_owned() {
-    if r_ns == primes::CORE_NS {
+  } else if let Some(ResolvedDef {
+    ns: r_ns,
+    def: _r_def,
+    rule: import_rule,
+  }) = resolved.to_owned()
+  {
+    if &*r_ns == primes::CORE_NS {
       // functions under core uses built $calcit module entry
       return Ok(format!("{}{}", var_prefix, escape_var(s)));
     }
-    if let Some(ImportRule::NsDefault(_s)) = import_rule {
+    if let Some(ImportRule::NsDefault(_s)) = import_rule.map(|x| (&*x).to_owned()) {
       // imports that using :default are special
-      track_ns_import(s.to_owned(), ImportedTarget::DefaultNs(r_ns), file_imports)?;
+      track_ns_import(s, ImportedTarget::DefaultNs(r_ns), file_imports)?;
     } else {
-      track_ns_import(s.to_owned(), ImportedTarget::ReferNs(r_ns), file_imports)?;
+      track_ns_import(s, ImportedTarget::ReferNs(r_ns), file_imports)?;
     }
     Ok(escape_var(s))
   } else if def_ns == primes::CORE_NS {
@@ -627,7 +636,7 @@ fn gen_symbol_code(
   } else if def_ns.is_empty() {
     Err(format!("Unexpected ns at symbol, {:?}", xs))
   } else if def_ns != ns {
-    track_ns_import(s.to_owned(), ImportedTarget::ReferNs(def_ns.to_owned()), file_imports)?;
+    track_ns_import(s, ImportedTarget::ReferNs(def_ns.into()), file_imports)?;
 
     // probably via macro
     // TODO dirty code collecting imports
@@ -643,25 +652,18 @@ fn gen_symbol_code(
 }
 
 // track but compare first, return Err if a different one existed
-fn track_ns_import(
-  sym: String,
-  import_rule: ImportedTarget,
-  file_imports: &RefCell<ImportsDict>,
-) -> Result<(), String> {
+fn track_ns_import(sym: &str, import_rule: ImportedTarget, file_imports: &RefCell<ImportsDict>) -> Result<(), String> {
   let mut dict = file_imports.borrow_mut();
-  match dict.get(&sym) {
+  match dict.get(&Arc::from(sym.to_owned())) {
     Some(v) => {
       if *v == import_rule {
         Ok(())
       } else {
-        Err(format!(
-          "conflicted import rule, previous {:?}, now {:?}",
-          v, import_rule
-        ))
+        Err(format!("conflicted import rule, previous {:?}, now {:?}", v, import_rule))
       }
     }
     None => {
-      dict.insert(sym, import_rule);
+      dict.insert(sym.to_owned().into(), import_rule);
       Ok(())
     }
   }
@@ -669,11 +671,11 @@ fn track_ns_import(
 
 fn gen_let_code(
   body: &CalcitItems,
-  local_defs: &HashSet<String>,
+  local_defs: &HashSet<Arc<str>>,
   xs: &Calcit,
   ns: &str,
   file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<String>>,
+  keywords: &RefCell<Vec<EdnKwd>>,
   base_return_label: &Option<String>,
 ) -> Result<String, String> {
   let mut let_def_body = body.to_owned();
@@ -693,13 +695,13 @@ fn gen_let_code(
       return Err(format!("&let expected body, but got empty, {}", xs.lisp_str()));
     }
     let pair = let_def_body[0].to_owned();
-    let content = let_def_body.skip(1);
+    let content = let_def_body.skip(1)?;
 
     match &pair {
       Calcit::Nil => {
         // non content defs_code
 
-        for (idx, x) in content.iter().enumerate() {
+        for (idx, x) in content.into_iter().enumerate() {
           if idx == content.len() - 1 {
             body_part.push_str(&to_js_code(x, ns, &scoped_defs, file_imports, keywords, &return_label)?);
             body_part.push('\n');
@@ -715,14 +717,14 @@ fn gen_let_code(
         let def_code = xs[1].to_owned();
 
         match def_name {
-          Calcit::Symbol(sym, ..) => {
+          Calcit::Symbol { sym, .. } => {
             // TODO `let` inside expressions makes syntax error
             let left = escape_var(&sym);
             let right = to_js_code(&def_code, ns, &scoped_defs, file_imports, keywords, &None)?;
             defs_code.push_str(&format!("let {} = {};\n", left, right));
 
             if scoped_defs.contains(&sym) {
-              for (idx, x) in content.iter().enumerate() {
+              for (idx, x) in content.into_iter().enumerate() {
                 if idx == content.len() - 1 {
                   body_part.push_str(&to_js_code(x, ns, &scoped_defs, file_imports, keywords, &return_label)?);
                   body_part.push('\n');
@@ -749,9 +751,9 @@ fn gen_let_code(
               if content.len() == 1 {
                 match &content[0] {
                   Calcit::List(ys) if ys.len() > 2 => match (&ys[0], &ys[1]) {
-                    (Calcit::Syntax(sym, _ns), Calcit::List(zs)) if sym == "&let" && zs.len() == 2 => match &zs[0] {
-                      Calcit::Symbol(s2, ..) if !scoped_defs.contains(s2) => {
-                        let_def_body = ys.skip(1);
+                    (Calcit::Syntax(sym, _ns), Calcit::List(zs)) if sym == &CalcitSyntax::CoreLet && zs.len() == 2 => match &zs[0] {
+                      Calcit::Symbol { sym: s2, .. } if !scoped_defs.contains(s2) => {
+                        let_def_body = ys.skip(1)?;
                         continue;
                       }
                       _ => (),
@@ -762,7 +764,7 @@ fn gen_let_code(
                 }
               }
 
-              for (idx, x) in content.iter().enumerate() {
+              for (idx, x) in content.into_iter().enumerate() {
                 if idx == content.len() - 1 {
                   body_part.push_str(&to_js_code(x, ns, &scoped_defs, file_imports, keywords, &return_label)?);
                   body_part.push('\n');
@@ -791,11 +793,11 @@ fn gen_let_code(
 
 fn gen_if_code(
   body: &CalcitItems,
-  local_defs: &HashSet<String>,
+  local_defs: &HashSet<Arc<str>>,
   _xs: &Calcit,
   ns: &str,
   file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<String>>,
+  keywords: &RefCell<Vec<EdnKwd>>,
   base_return_label: &Option<String>,
 ) -> Result<String, String> {
   if body.len() < 2 || body.len() > 3 {
@@ -822,7 +824,7 @@ fn gen_if_code(
       if let Some(false_node) = some_false_node {
         if let Calcit::List(ys) = false_node {
           if let Some(Calcit::Syntax(syn, _ns)) = ys.get(0) {
-            if syn == "if" {
+            if syn == &CalcitSyntax::If {
               if ys.len() < 3 || ys.len() > 4 {
                 return Err(format!("if expected 2~3 nodes, got: {:?}", ys));
               }
@@ -837,11 +839,10 @@ fn gen_if_code(
 
         let false_code = to_js_code(false_node, ns, local_defs, file_imports, keywords, &return_label)?;
         chunk.push_str(&format!("else {{ {} }}", false_code));
-        break;
       } else {
         chunk.push_str(&format!("else {{ {} null; }}", return_label.unwrap()));
-        break;
       }
+      break;
     }
 
     if base_return_label.is_some() {
@@ -855,16 +856,16 @@ fn gen_if_code(
 fn gen_args_code(
   body: &CalcitItems,
   ns: &str,
-  local_defs: &HashSet<String>,
+  local_defs: &HashSet<Arc<str>>,
   file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<String>>,
+  keywords: &RefCell<Vec<EdnKwd>>,
 ) -> Result<String, String> {
   let mut result = String::from("");
   let var_prefix = if ns == "calcit.core" { "" } else { "$calcit." };
   let mut spreading = false;
   for x in body {
     match x {
-      Calcit::Symbol(s, ..) if s == "&" => {
+      Calcit::Symbol { sym, .. } if &**sym == "&" => {
         spreading = true;
       }
       _ => {
@@ -890,24 +891,17 @@ fn gen_args_code(
 fn list_to_js_code(
   xs: &CalcitItems,
   ns: &str,
-  local_defs: HashSet<String>,
+  local_defs: HashSet<Arc<str>>,
   return_label: &str,
   file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<String>>,
+  keywords: &RefCell<Vec<EdnKwd>>,
 ) -> Result<String, String> {
   // TODO default returnLabel="return "
   let mut result = String::from("");
-  for (idx, x) in xs.iter().enumerate() {
+  for (idx, x) in xs.into_iter().enumerate() {
     // result = result & "// " & $x & "\n"
     if idx == xs.len() - 1 {
-      let line = to_js_code(
-        x,
-        ns,
-        &local_defs,
-        file_imports,
-        keywords,
-        &Some(return_label.to_owned()),
-      )?;
+      let line = to_js_code(x, ns, &local_defs, file_imports, keywords, &Some(return_label.to_owned()))?;
       result.push_str(&line);
       result.push('\n');
     } else {
@@ -925,11 +919,11 @@ fn list_to_js_code(
 
 fn uses_recur(xs: &Calcit) -> bool {
   match xs {
-    Calcit::Symbol(s, ..) => s == "recur",
-    Calcit::Proc(s) => s == "recur",
+    Calcit::Symbol { sym: s, .. } => &**s == "recur",
+    Calcit::Proc(s) => &**s == "recur",
     Calcit::List(ys) => match &ys.get(0) {
-      Some(Calcit::Syntax(syn, _)) if syn == "defn" => false,
-      Some(Calcit::Symbol(sym, ..)) if sym == "defn" => false,
+      Some(Calcit::Syntax(syn, _)) if syn == &CalcitSyntax::Defn => false,
+      Some(Calcit::Symbol { sym, .. }) if &**sym == "defn" => false,
       _ => {
         for y in ys {
           if uses_recur(y) {
@@ -949,9 +943,9 @@ fn gen_js_func(
   raw_body: &CalcitItems,
   ns: &str,
   exported: bool,
-  outer_defs: &HashSet<String>,
+  outer_defs: &HashSet<Arc<str>>,
   file_imports: &RefCell<ImportsDict>,
-  keywords: &RefCell<Vec<String>>,
+  keywords: &RefCell<Vec<EdnKwd>>,
 ) -> Result<String, String> {
   let var_prefix = if ns == "calcit.core" { "" } else { "$calcit." };
   let mut local_defs = outer_defs.to_owned();
@@ -963,7 +957,7 @@ fn gen_js_func(
   let mut optional_count = 0;
   for x in args {
     match x {
-      Calcit::Symbol(sym, ..) => {
+      Calcit::Symbol { sym, .. } => {
         if spreading {
           if !args_code.is_empty() {
             args_code.push_str(", ");
@@ -983,11 +977,11 @@ fn gen_js_func(
           args_code.push_str(&escape_var(sym));
           optional_count += 1;
         } else {
-          if sym == "&" {
+          if &**sym == "&" {
             spreading = true;
             continue;
           }
-          if sym == "?" {
+          if &**sym == "?" {
             has_optional = true;
             continue;
           }
@@ -1011,13 +1005,13 @@ fn gen_js_func(
     snippets::tmpl_args_exact(args_count)
   };
 
-  let mut body: CalcitItems = im::Vector::new();
+  let mut body: CalcitItems = TernaryTreeList::Empty;
   let mut async_prefix: String = String::from("");
 
   for line in raw_body {
     if let Calcit::List(xs) = line {
       if let Some(Calcit::Syntax(sym, _ns)) = xs.get(0) {
-        if sym == "hint-fn" {
+        if sym == &CalcitSyntax::HintFn {
           if hinted_async(xs) {
             async_prefix = String::from("async ")
           }
@@ -1025,19 +1019,12 @@ fn gen_js_func(
         }
       }
     }
-    body.push_back(line.to_owned());
+    body = body.push(line.to_owned());
   }
 
   if !body.is_empty() && uses_recur(&body[body.len() - 1]) {
     let return_var = js_gensym("return_mark");
-    let body = list_to_js_code(
-      &body,
-      ns,
-      local_defs,
-      &format!("%%{}%% =", return_var),
-      file_imports,
-      keywords,
-    )?;
+    let body = list_to_js_code(&body, ns, local_defs, &format!("%%{}%% =", return_var), file_imports, keywords)?;
     let fn_def = snippets::tmpl_tail_recursion(
       /* name = */ escape_var(name),
       /* args_code = */ args_code,
@@ -1072,10 +1059,10 @@ fn gen_js_func(
 }
 
 /// this is a very rough implementation for now
-fn hinted_async(xs: &im::Vector<Calcit>) -> bool {
+fn hinted_async(xs: &TernaryTreeList<Calcit>) -> bool {
   for x in xs {
     match x {
-      Calcit::Symbol(sym, ..) if sym == "async" => return true,
+      Calcit::Symbol { sym, .. } if &**sym == "async" => return true,
       _ => {}
     }
   }
@@ -1084,10 +1071,10 @@ fn hinted_async(xs: &im::Vector<Calcit>) -> bool {
 
 fn contains_symbol(xs: &Calcit, y: &str) -> bool {
   match xs {
-    Calcit::Symbol(s, ..) => s == y,
+    Calcit::Symbol { sym, .. } => &**sym == y,
     Calcit::Thunk(code, _) => contains_symbol(code, y),
-    Calcit::Fn(_, _, _, _, _, body) => {
-      for x in body {
+    Calcit::Fn { body, .. } => {
+      for x in &**body {
         if contains_symbol(x, y) {
           return true;
         }
@@ -1106,12 +1093,12 @@ fn contains_symbol(xs: &Calcit, y: &str) -> bool {
   }
 }
 
-fn sort_by_deps(deps: &HashMap<String, Calcit>) -> Vec<String> {
-  let mut deps_graph: HashMap<String, HashSet<String>> = HashMap::new();
-  let mut def_names: Vec<String> = vec![];
+fn sort_by_deps(deps: &HashMap<Arc<str>, Calcit>) -> Vec<Arc<str>> {
+  let mut deps_graph: HashMap<Arc<str>, HashSet<Arc<str>>> = HashMap::new();
+  let mut def_names: Vec<Arc<str>> = Vec::with_capacity(deps.len());
   for (k, v) in deps {
     def_names.push(k.to_owned());
-    let mut deps_info: HashSet<String> = HashSet::new();
+    let mut deps_info: HashSet<Arc<str>> = HashSet::new();
     for k2 in deps.keys() {
       if k2 == k {
         continue;
@@ -1126,7 +1113,7 @@ fn sort_by_deps(deps: &HashMap<String, Calcit>) -> Vec<String> {
   // println!("\ndefs graph {:?}", deps_graph);
   def_names.sort(); // alphabet order first
 
-  let mut result: Vec<String> = vec![];
+  let mut result: Vec<Arc<str>> = Vec::with_capacity(def_names.len());
   'outer: for x in def_names {
     for (idx, y) in result.iter().enumerate() {
       if depends_on(y, &x, &deps_graph, 3) {
@@ -1142,12 +1129,12 @@ fn sort_by_deps(deps: &HashMap<String, Calcit>) -> Vec<String> {
 }
 
 // could be slow, need real topology sorting
-fn depends_on(x: &str, y: &str, deps: &HashMap<String, HashSet<String>>, decay: usize) -> bool {
+fn depends_on(x: &str, y: &str, deps: &HashMap<Arc<str>, HashSet<Arc<str>>>, decay: usize) -> bool {
   if decay == 0 {
     false
   } else {
     for item in &deps[x] {
-      if item == y || depends_on(item, y, deps, decay - 1) {
+      if &**item == y || depends_on(item, y, deps, decay - 1) {
         return true;
       } else {
         // nothing
@@ -1171,7 +1158,7 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
     let _ = fs::create_dir(code_emit_path);
   }
 
-  let mut unchanged_ns: HashSet<String> = HashSet::new();
+  let mut unchanged_ns: HashSet<Arc<str>> = HashSet::new();
 
   let program = program::clone_evaled_program();
   for (ns, file) in program {
@@ -1179,9 +1166,9 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
     // side-effects, reset tracking state
 
     let file_imports: RefCell<ImportsDict> = RefCell::new(BTreeMap::new());
-    let keywords: RefCell<Vec<String>> = RefCell::new(Vec::new());
+    let keywords: RefCell<Vec<EdnKwd>> = RefCell::new(Vec::new());
 
-    let mut defs_in_current: HashSet<String> = HashSet::new();
+    let mut defs_in_current: HashSet<Arc<str>> = HashSet::new();
     for k in file.keys() {
       defs_in_current.insert(k.to_owned());
     }
@@ -1213,13 +1200,13 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
     let mut direct_code = String::from(""); // dirty code to run directly
     let mut keywords_code = String::from("\nvar _kwd={};"); // initialization for keywords
 
-    let mut import_code = if ns == "calcit.core" {
+    let mut import_code = if &*ns == "calcit.core" {
       snippets::tmpl_import_procs(wrap_js_str("@calcit/procs"))
     } else {
       format!("\nimport * as $calcit from {};", core_lib)
     };
 
-    let mut def_names: HashSet<String> = HashSet::new(); // multiple parts of scoped defs need to be tracked
+    let mut def_names: HashSet<Arc<str>> = HashSet::new(); // multiple parts of scoped defs need to be tracked
 
     // tracking top level scope definitions
     for def in file.keys() {
@@ -1230,17 +1217,13 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
     // println!("deps order: {:?}", deps_in_order);
 
     for def in deps_in_order {
-      if ns == primes::CORE_NS {
+      if &*ns == primes::CORE_NS {
         // some defs from core can be replaced by calcit.procs
         if is_js_unavailable_procs(&def) {
           continue;
         }
         if is_preferred_js_proc(&def) {
-          defs_code.push_str(&format!(
-            "\nvar {} = $calcit_procs.{};\n",
-            escape_var(&def),
-            escape_var(&def)
-          ));
+          defs_code.push_str(&format!("\nvar {} = $calcit_procs.{};\n", escape_var(&def), escape_var(&def)));
           continue;
         }
       }
@@ -1250,38 +1233,31 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
       match &f {
         // probably not work here
         Calcit::Proc(..) => {
-          defs_code.push_str(&format!(
-            "\nvar {} = $calcit_procs.{};\n",
-            escape_var(&def),
-            escape_var(&def)
-          ));
+          defs_code.push_str(&format!("\nvar {} = $calcit_procs.{};\n", escape_var(&def), escape_var(&def)));
         }
-        Calcit::Fn(name, def_ns, _, _, args, code) => {
-          call_stack::push_call_stack(def_ns, name, StackKind::Codegen, f.to_owned(), &im::vector![]);
-          defs_code.push_str(&gen_js_func(
-            &def,
-            args,
-            code,
-            &ns,
-            true,
-            &def_names,
-            &file_imports,
-            &keywords,
-          )?);
-          call_stack::pop_call_stack();
+        Calcit::Fn {
+          name,
+          def_ns,
+          args,
+          body: code,
+          ..
+        } => {
+          gen_stack::push_call_stack(def_ns, name, StackKind::Codegen, f.to_owned(), &TernaryTreeList::Empty);
+          defs_code.push_str(&gen_js_func(&def, args, code, &ns, true, &def_names, &file_imports, &keywords)?);
+          gen_stack::pop_call_stack();
         }
         Calcit::Thunk(code, _) => {
           // TODO need topological sorting for accuracy
           // values are called directly, put them after fns
-          call_stack::push_call_stack(&ns, &def, StackKind::Codegen, (**code).to_owned(), &im::vector![]);
+          gen_stack::push_call_stack(&ns, &def, StackKind::Codegen, (**code).to_owned(), &TernaryTreeList::Empty);
           vals_code.push_str(&format!(
             "\nexport var {} = {};\n",
             escape_var(&def),
             to_js_code(code, &ns, &def_names, &file_imports, &keywords, &None)?
           ));
-          call_stack::pop_call_stack()
+          gen_stack::pop_call_stack()
         }
-        Calcit::Macro(..) => {
+        Calcit::Macro { .. } => {
           // macro should be handled during compilation, psuedo code
           defs_code.push_str(&snippets::tmpl_export_macro(escape_var(&def)));
         }
@@ -1296,7 +1272,7 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
         }
       }
     }
-    if ns == primes::CORE_NS {
+    if &*ns == primes::CORE_NS {
       // add at end of file to register builtin classes
       direct_code.push_str(&snippets::tmpl_classes_registering())
     }
@@ -1314,11 +1290,7 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
               import_code.push_str(&format!("\nimport * as {} from {};", escape_ns(&def), import_target));
             } else {
               let import_target = to_js_import_name(&target_ns, false); // TODO js_mode
-              import_code.push_str(&format!(
-                "\nimport * as {} from {};",
-                escape_ns(&target_ns),
-                import_target
-              ));
+              import_code.push_str(&format!("\nimport * as {} from {};", escape_ns(&target_ns), import_target));
             }
           }
           ImportedTarget::DefaultNs(target_ns) => {
@@ -1341,7 +1313,7 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
       }
     }
 
-    let kwd_prefix = if ns == "calcit.core" { "" } else { "$calcit." };
+    let kwd_prefix = if &*ns == "calcit.core" { "" } else { "$calcit." };
     let mut kwd_arr = String::from("[");
     {
       // need to maintain a stable order to reduce redundant reloads
@@ -1349,7 +1321,7 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
       kwds.dedup();
     }
     for s in keywords.into_inner() {
-      let name = escape_cirru_str(&s);
+      let name = escape_cirru_str(&s.to_string());
       kwd_arr.push_str(&format!("{},", name));
     }
     kwd_arr.push(']');
@@ -1359,10 +1331,7 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
     let js_file_path = code_emit_path.join(to_js_file_name(&ns, false)); // TODO mjs_mode
     let wrote_new = write_file_if_changed(
       &js_file_path,
-      &format!(
-        "{}\n{}{}\n\n{}\n{}",
-        import_code, keywords_code, defs_code, vals_code, direct_code
-      ),
+      &format!("{}\n{}{}\n\n{}\n{}", import_code, keywords_code, defs_code, vals_code, direct_code),
     );
     if wrote_new {
       println!("Emitted js file: {}", js_file_path.to_str().unwrap());
@@ -1383,18 +1352,20 @@ pub fn emit_js(entry_ns: &str, emit_path: &str) -> Result<(), String> {
 fn is_js_unavailable_procs(name: &str) -> bool {
   matches!(
     name,
-    "&reset-gensym-index!"
-      | "dbt->point"
-      | "dbt-digits" // TODO none
-      | "dbt-balanced-ternary"
-      | "gensym"
-      | "macroexpand"
-      | "macroexpand-all"
-      | "to-cirru-edn"
-      | "extract-cirru-edn"
+    "&reset-gensym-index!" | "gensym" | "macroexpand" | "macroexpand-all" | "to-cirru-edn" | "extract-cirru-edn"
   )
 }
 
+#[inline(always)]
 fn is_cirru_string(s: &str) -> bool {
   s.starts_with('|') || s.starts_with('"')
 }
+
+#[inline(always)]
+fn get_proc_prefix(ns: &str) -> &str {
+  if ns == primes::CORE_NS {
+    "$calcit_procs."
+  } else {
+    "$calcit."
+  }
+}
